#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'tianqin'

import os
import sys
import logging
import threading
import tempfile
import argparse
import datetime
import traceback
from functools import partial


class LogHandlerChan(logging.Handler):
    def __init__(self, chan, dt_func):
        logging.Handler.__init__(self)
        self.chan = chan
        self.dt_func = dt_func

    def emit(self, record):
        d = {
            "aid": "log",
            "datetime": self.dt_func(),
            "level": str(record.levelname),
            "content": record.msg % record.args,
        }
        if record.exc_info:
            ls = traceback.format_exception(record.exc_info[0], record.exc_info[1], record.exc_info[2])
            d["content"] += "".join(ls)
        self.chan.send_nowait(d)


class PrintWriterToLog:
    def __init__(self, logger):
        self.orign_stdout = sys.stdout
        self.line = ""
        self.logger = logger

    def write(self, text):
        if self.orign_stdout:
            self.orign_stdout.write(text)
        self.line += text
        if self.line[-1] == "\n":
            self.logger.info(self.line[:-1])
            self.line = ""

    def flush(self):
        if self.orign_stdout:
            self.orign_stdout.flush()


def exception_handler(api, orign_hook, type, value, tb):
    logging.getLogger("TQ").error("程序异常", exc_info=(type, value, tb))
    api.close()
    orign_hook(type, value, tb)


def write_snapshot(dt_func, tq_send_chan, account, positions):
    tq_send_chan.send_nowait({
        "aid": "snapshot",
        "datetime": dt_func(),
        "accounts": {
            "CNY": {k: v for k, v in account.items() if not k.startswith("_")},
        },
        "positions": {k: {pk: pv for pk, pv in v.items() if not pk.startswith("_")} for k, v in positions.items() if
                      not k.startswith("_")},
    })


async def account_watcher(api, dt_func, tq_send_chan):
    account = api.get_account()
    positions = api.get_position()
    try:
        async with api.register_update_notify() as update_chan:
            async for _ in update_chan:
                account_changed = api.is_changing(account, "static_balance")
                for d in api._diffs:
                    for oid in d.get("trade", {}).get(api._account._account_id, {}).get("orders", {}).keys():
                        order = api.get_order(oid)
                        if order._this_session is not True:
                            continue
                        account_changed = True
                        tq_send_chan.send_nowait({
                            "aid": "order",
                            "datetime": dt_func(),
                            "order": {k: v for k, v in order.items() if not k.startswith("_")},
                        })
                    for tid in d.get("trade", {}).get(api._account._account_id, {}).get("trades", {}).keys():
                        trade = api.get_trade(tid)
                        order = api.get_order(trade.order_id)
                        if order._this_session is not True:
                            continue
                        account_changed = True
                        tq_send_chan.send_nowait({
                            "aid": "trade",
                            "datetime": dt_func(),
                            "trade": {k: v for k, v in trade.items() if not k.startswith("_")},
                        })
                if account_changed:
                    write_snapshot(dt_func, tq_send_chan, account, positions)
    finally:
        write_snapshot(dt_func, tq_send_chan, account, positions)


class TqMonitorThread(threading.Thread):
    """
    监控天勤进程存活情况
    """

    def __init__(self, tq_pid):
        threading.Thread.__init__(self, daemon=True)
        self.tq_pid = tq_pid

    def run(self):
        # TODO: 如果发布 mac / linux 天勤客户端，未来还需要在 mac / linux 上实现同样的功能。
        if not sys.platform.startswith("win"):
            return
        import _winapi
        try:
            p = _winapi.OpenProcess(_winapi.PROCESS_ALL_ACCESS, False, self.tq_pid)
            os.waitpid(p, 0)
        except:
            pass
        finally:
            os._exit(0)


if sys.platform != "win32":
    import fcntl


class SingleInstanceException(BaseException):
    pass


def get_self_full_name():
    s = os.path.abspath(sys.argv[0])
    return s[0].upper() + s[1:]


class SingleInstance(object):
    def __init__(self, flavor_id=""):
        self.initialized = False
        self.instance_id = get_self_full_name().replace(
            "/", "-").replace(":", "").replace("\\", "-") + '-%s' % flavor_id
        self.lockfile = os.path.normpath(
            tempfile.gettempdir() + '/' + self.instance_id + '.lock')

        if sys.platform == 'win32':
            try:
                # file already exists, we try to remove (in case previous
                # execution was interrupted)
                if os.path.exists(self.lockfile):
                    os.unlink(self.lockfile)
                self.fd = os.open(
                    self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
            except OSError:
                type, e, tb = sys.exc_info()
                if e.errno == 13:
                    raise SingleInstanceException()
                print(e.errno)
                raise
        else:  # non Windows
            self.fp = open(self.lockfile, 'w')
            self.fp.flush()
            try:
                fcntl.lockf(self.fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError:
                raise SingleInstanceException()
        self.initialized = True

    def __del__(self):
        if not self.initialized:
            return
        try:
            if sys.platform == 'win32':
                if hasattr(self, 'fd'):
                    os.close(self.fd)
                    os.unlink(self.lockfile)
            else:
                fcntl.lockf(self.fp, fcntl.LOCK_UN)
                if os.path.isfile(self.lockfile):
                    os.unlink(self.lockfile)
        except Exception as e:
            sys.exit(-1)


class Forwarding(object):
    """
    作为数据转发的中间模块, 创建与天勤的新连接, 并过滤TqApi向上游发送的数据.
    对于这个新连接: 发送所有的 set_chart_data 类型数据包,以及抄送所有的 subscribe_quote, set_chart类型数据包
    """

    def __init__(self, api, api_send_chan, api_recv_chan, upstream_send_chan, upstream_recv_chan, tq_send_chan,
                 tq_recv_chan):
        self.api = api
        self.api_send_chan = api_send_chan
        self.api_recv_chan = api_recv_chan
        self.upstream_send_chan = upstream_send_chan
        self.upstream_recv_chan = upstream_recv_chan
        self.tq_send_chan = tq_send_chan
        self.tq_recv_chan = tq_recv_chan
        self.subscribed = None
        self.order_symbols = set()

    async def _forward(self):
        to_downstream_task = self.api.create_task(
            self._forward_to_downstream(self.api_recv_chan, self.upstream_recv_chan))  # 转发给下游
        to_upstream_task = self.api.create_task(
            self._forward_to_upstream(self.api_send_chan, self.upstream_send_chan, self.tq_send_chan))  # 转发给上游
        try:
            async for pack in self.tq_recv_chan:
                pass
        finally:
            to_downstream_task.cancel()
            to_upstream_task.cancel()

    async def _forward_to_downstream(self, api_recv_chan, upstream_recv_chan):
        """转发给下游"""
        async for pack in upstream_recv_chan:
            await api_recv_chan.send(pack)

    async def _forward_to_upstream(self, api_send_chan, upstream_send_chan, tq_send_chan):
        """转发给上游"""
        async for pack in api_send_chan:
            if pack["aid"] == "set_chart_data":
                await tq_send_chan.send(pack)
            elif pack["aid"] == "insert_order":
                self.order_symbols.add(pack["exchange_id"] + "." + pack["instrument_id"])
                await self._send_subscribed_to_tq()
            elif pack["aid"] == "set_chart" or pack["aid"] == "subscribe_quote":
                await self._send_subscribed_to_tq()
            # 所有数据都需要向上游（新版tqwebhelper）转发，由上游判断是否继续转发
            await upstream_send_chan.send(pack)

    async def _send_subscribed_to_tq(self):
        d = []
        for item in self.api._requests["klines"].keys():
            for symbol in item[0]:  # 如果同时订阅多个合约，分别发送给tq
                d.append({
                    "symbol": symbol,
                    "dur_nano": item[1] * 1000000000
                })
        for item in self.api._requests["ticks"].keys():
            d.append({
                "symbol": item[0],
                "dur_nano": 0
            })
        for symbol in self.api._requests["quotes"]:
            d.append({
                "symbol": symbol
            })
        for symbol in self.order_symbols:
            d.append({
                "symbol": symbol
            })
        if d != self.subscribed:
            self.subscribed = d
            await self.tq_send_chan.send({
                "aid": "subscribed",
                "subscribed": self.subscribed
            })


def link_tq(api):
    """
    处理py进程到天勤的连接

    根据天勤提供的命令行参数, 决定 TqApi 工作方式

    * 直接调整api的参数

    TqApi运行过程中的一批信息主动发送到天勤显示

    * 进程启动和停止
    * set_chart_data 指令全部发往天勤绘图
    * 所有 log / print 信息传递一份
    * exception 发送一份
    * 所有报单/成交记录抄送一份

    :return: (account, backtest, md_url)
    """
    from tqsdk.api import TqChan, TqAccount
    from tqsdk.sim import TqSim
    # 解析命令行参数
    parser = argparse.ArgumentParser()
    # 天勤连接基本参数
    parser.add_argument('--_action', type=str, required=False)
    parser.add_argument('--_tq_pid', type=int, required=False)
    parser.add_argument('--_tq_url', type=str, required=False)
    # action==run时需要这几个
    parser.add_argument('--_broker_id', type=str, required=False)
    parser.add_argument('--_account_id', type=str, required=False)
    parser.add_argument('--_password', type=str, required=False)
    # action==backtest时需要这几个
    parser.add_argument('--_start_dt', type=str, required=False)
    parser.add_argument('--_end_dt', type=str, required=False)
    parser.add_argument('--_init_balance', type=str, required=False)
    # action==mdreplay时需要这几个
    parser.add_argument('--_ins_url', type=str, required=False)
    parser.add_argument('--_md_url', type=str, required=False)
    args, unknown = parser.parse_known_args()

    # 非天勤启动时直接返回
    if args._action is None:
        return None, None
    if args._tq_pid is None:
        raise Exception("_tq_pid 参数缺失")
    if args._tq_url is None:
        raise Exception("_tq_url 参数缺失")
    if args._action == "run" and (not args._broker_id or not args._account_id or not args._password):
        raise Exception("run 必要参数缺失")
    if args._action == "backtest" and (not args._start_dt or not args._end_dt):
        raise Exception("backtest 必要参数缺失")
    if args._action == "mdreplay" and (not args._ins_url or not args._md_url):
        raise Exception("mdreplay 必要参数缺失")

    # 监控天勤进程存活情况
    TqMonitorThread(args._tq_pid).start()

    # 建立到天勤进程的连接
    tq_send_chan, tq_recv_chan = TqChan(api), TqChan(api)  # 连接到天勤的channel
    api.create_task(api._connect(args._tq_url, tq_send_chan, tq_recv_chan))  # 启动到天勤客户端的连接

    # 根据运行模式分别执行不同的初始化任务
    if args._action == "run":
        if isinstance(api._account, TqAccount) and (
                api._account._broker_id != args._broker_id or api._account._account_id != args._account_id):
            raise Exception("策略代码与设置中的账户参数冲突。可尝试删去代码中的账户参数 TqAccount，以终端或者插件设置的账户参数运行。")
        instance = SingleInstance(args._account_id)
        api._account = TqAccount(args._broker_id, args._account_id, args._password)
        api._backtest = None
        dt_func = lambda: int(datetime.datetime.now().timestamp() * 1e9)
        tq_send_chan.send_nowait({
            "aid": "register_instance",
            "instance_id": instance.instance_id,
            "full_path": get_self_full_name(),
            "instance_pid": os.getpid(),
            "instance_type": "RUN",
            "broker_id": args._broker_id,
            "account_id": args._account_id,
            "password": args._password,
        })
    elif args._action == "backtest":
        instance = SingleInstance("%s-%s" % (args._start_dt, args._end_dt))
        if args._init_balance:
            try:
                api._account = TqSim(float(args._init_balance))
            except ValueError:
                raise Exception("backtest 参数错误, _init_balance = " + args._init_balance + " 不是数字")
        elif not isinstance(api._account, TqSim):
            api._account = TqSim()
        from tqsdk.backtest import TqBacktest
        start_date = datetime.datetime.strptime(args._start_dt, '%Y%m%d')
        end_date = datetime.datetime.strptime(args._end_dt, '%Y%m%d')
        api._backtest = TqBacktest(start_dt=start_date, end_dt=end_date)
        dt_func = lambda: api._account._get_current_timestamp()
        tq_send_chan.send_nowait({
            "aid": "register_instance",
            "instance_id": instance.instance_id,
            "full_path": get_self_full_name(),
            "instance_pid": os.getpid(),
            "instance_type": "BACKTEST",
            "start_dt": args._start_dt,
            "end_dt": args._end_dt,
        })
    elif args._action == "mdreplay":
        instance = SingleInstance(args._account_id)
        api._account = TqSim(account_id=args._account_id)
        api._backtest = None
        api._md_url = args._md_url
        api._ins_url = args._ins_url
        dt_func = lambda: api._account._get_current_timestamp()
        tq_send_chan.send_nowait({
            "aid": "register_instance",
            "instance_id": instance.instance_id,
            "full_path": get_self_full_name(),
            "instance_pid": os.getpid(),
            "instance_type": "REPLAY",
            "account_id": "SIM",
        })
    else:
        raise Exception("_action 参数异常")

    # print输出, exception信息转发到天勤
    logger = logging.getLogger("TQ")
    logger.setLevel(logging.INFO)
    logger.addHandler(LogHandlerChan(tq_send_chan, dt_func=dt_func))  # log输出到天勤接口
    sys.stdout = PrintWriterToLog(logger)  # print信息转向log输出
    sys.excepthook = partial(exception_handler, api, sys.excepthook)  # exception信息转向log输出

    # 向api注入监控任务, 将账户交易信息主动推送到天勤
    api.create_task(account_watcher(api, dt_func, tq_send_chan))
    return tq_send_chan, tq_recv_chan



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

import json
from datetime import datetime
from tqsdk import TqApi, TqBacktest


def test_get_quote_normal_backtest():
    """
    回测获取行情报价

    """
    # 获取行情
    api = TqApi(backtest=TqBacktest(datetime(2019, 10, 15), datetime(2019, 10, 16)),
                _ins_url="https://openmd.shinnytech.com/t/md/symbols/2019-07-03.json")
    quote = api.get_quote("SHFE.cu2001")
    quote_data = {k: v for k, v in quote.items()}
    quote_data["trading_time"] = {k: v for k, v in quote_data["trading_time"].items()}

    assert json.dumps(quote_data, sort_keys=True) == \
           '{"amount": NaN, "ask_price1": 47070.0, "ask_price2": NaN, "ask_price3": NaN, "ask_price4": NaN, "ask_price5": NaN, "ask_volume1": 1, "ask_volume2": 0, "ask_volume3": 0, "ask_volume4": 0, "ask_volume5": 0, "average": NaN, "bid_price1": 47050.0, "bid_price2": NaN, "bid_price3": NaN, "bid_price4": NaN, "bid_price5": NaN, "bid_volume1": 1, "bid_volume2": 0, "bid_volume3": 0, "bid_volume4": 0, "bid_volume5": 0, "close": NaN, "commission": 11.594999999999999, "datetime": "2019-10-14 23:59:59.999999", "delivery_month": 1, "delivery_year": 2020, "expire_datetime": 1579071600.0, "expired": false, "highest": NaN, "ins_class": "FUTURE", "instrument_id": "SHFE.cu2001", "last_price": 47060.0, "lower_limit": NaN, "lowest": NaN, "margin": 16233.000000000002, "max_limit_order_volume": 500, "max_market_order_volume": 0, "min_limit_order_volume": 0, "min_market_order_volume": 0, "open": NaN, "open_interest": 90714, "pre_close": NaN, "pre_open_interest": 0, "pre_settlement": NaN, "price_decs": 0, "price_tick": 10, "settlement": NaN, "strike_price": NaN, "trading_time": {"day": [["09:00:00", "10:15:00"], ["10:30:00", "11:30:00"], ["13:30:00", "15:00:00"]], "night": [["21:00:00", "25:00:00"]]}, "underlying_symbol": "", "upper_limit": NaN, "volume": 0, "volume_multiple": 5}'

    # 其他取值方式
    assert quote["pre_close"] != quote.pre_close
    assert quote.get("pre_settlement") != quote.pre_settlement
    assert quote.get("highest") != quote.highest
    assert quote.get("lowest") != quote.lowest
    assert quote["open"] != quote.open
    assert quote["close"] != quote.close
    api.close()



#!usr/bin/env python3
#-*- coding:utf-8 -*-
"""
@author: yanqiong
@file: guihelper.py
@create_on: 2019/11/7
@description: 
"""
import os
import sys
import argparse
import simplejson
import asyncio
from datetime import datetime
from aiohttp import web
import socket
import websockets
import tqsdk

class TqWebHelper(object):
    async def _run(self, api, api_send_chan, api_recv_chan, web_send_chan, web_recv_chan):
        self._api = api
        self._logger = self._api._logger.getChild("TqWebHelper")  # 调试信息输出
        if not self._api._web_gui:
            # 没有开启 web_gui 功能
            _data_handler_without_web_task = self._api.create_task(
                self._data_handler_without_web(api_recv_chan, web_recv_chan))
            try:
                async for pack in api_send_chan:
                    # api 发送的包，过滤出 set_chart_data, set_web_chart_data, 其余的原样转发
                    if pack['aid'] != 'set_chart_data' and pack['aid'] != 'set_web_chart_data':
                        await web_send_chan.send(pack)
            finally:
                _data_handler_without_web_task.cancel()
        else:
            # 解析命令行参数
            parser = argparse.ArgumentParser()
            # 天勤连接基本参数
            parser.add_argument('--_http_server_port', type=int, required=False)
            args, unknown = parser.parse_known_args()
            # 可选参数，tqwebhelper 中 http server 的 port
            if args._http_server_port is not None:
                self._http_server_port = args._http_server_port
            else:
                self._http_server_port = 0

            self._web_dir = os.path.join(os.path.dirname(__file__), 'web')
            file_path = os.path.abspath(sys.argv[0])
            file_name = os.path.basename(file_path)
            # 初始化数据截面
            self._data = {
                "action": {
                    "mode": "run",
                    "md_url_status": '-',
                    "td_url_status": True if isinstance(self._api._account, tqsdk.api.TqSim) else '-',
                    "account_id": self._api._account._account_id,
                    "broker_id": self._api._account._broker_id if isinstance(self._api._account, tqsdk.api.TqAccount) else 'TQSIM',
                    "file_path": file_path[0].upper() + file_path[1:],
                    "file_name": file_name
                },
                "trade": {},
                "subscribed": [],
                "draw_chart_datas": {},
                "snapshots": {}
            }
            self._order_symbols = set()
            self._diffs = []
            self._conn_diff_chans = set()
            self.web_port_chan = tqsdk.api.TqChan(self._api)  # 记录到 ws port 的channel
            _data_task = self._api.create_task(self._data_handler(api_recv_chan, web_recv_chan))
            _wsserver_task = self._api.create_task(self.link_wsserver())
            _httpserver_task = self._api.create_task(self.link_httpserver())

            try:
                # api 发送的包，过滤出需要的包记录在 self._data
                async for pack in api_send_chan:
                    if pack['aid'] == 'set_chart_data' or pack['aid'] == 'set_web_chart_data':
                        # 发送的是绘图数据
                        # 旧版 tqhelper aid=set_chart_data，发送除 KSERIAL/SERIAL 之外的序列，因为其 KSERIAL/SERIAL 序列不符合 diff 协议
                        # 新版 tqwebhelper aid=set_web_chart_data 中发送的 KSERIAL/SERIAL 数据
                        diff_data = {}  # 存储 pack 中的 diff 数据的对象
                        for series_id, series in pack['datas'].items():
                            if (pack['aid'] == 'set_chart_data' and series["type"] != "KSERIAL" and series["type"] != "SERIAL") or\
                                    pack['aid'] == 'set_web_chart_data' :
                                diff_data[series_id] = series
                        if diff_data != {}:
                            web_diff = {'draw_chart_datas': {}}
                            web_diff['draw_chart_datas'][pack['symbol']] = {}
                            web_diff['draw_chart_datas'][pack['symbol']][pack['dur_nano']] = diff_data
                            TqWebHelper.merge_diff(self._data, web_diff)
                            for chan in self._conn_diff_chans:
                                self.send_to_conn_chan(chan, [web_diff])
                    else:
                        if pack["aid"] == "insert_order":
                            self._order_symbols.add(pack["exchange_id"] + "." + pack["instrument_id"])
                        if pack['aid'] == 'subscribe_quote' or pack["aid"] == "set_chart" or pack["aid"] == 'insert_order':
                            web_diff = {'subscribed': []}
                            for item in self._api._requests["klines"].keys():
                                web_diff['subscribed'].append({"symbol": item[0], "dur_nano": item[1] * 1000000000})
                            for item in self._api._requests["ticks"].keys():
                                web_diff['subscribed'].append({"symbol": item[0], "dur_nano": 0})
                            for symbol in self._api._requests["quotes"]:
                                web_diff['subscribed'].append({"symbol": symbol})
                            for symbol in self._order_symbols:
                                web_diff['subscribed'].append({"symbol": symbol})
                            if web_diff['subscribed'] != self._data['subscribed']:
                                self._data['subscribed'] = web_diff['subscribed']
                            for chan in self._conn_diff_chans:
                                self.send_to_conn_chan(chan, [web_diff])
                        # 发送的转发给上游
                        await web_send_chan.send(pack)
            finally:
                _data_task.cancel()
                _wsserver_task.cancel()
                _httpserver_task.cancel()

    async def _data_handler_without_web(self, api_recv_chan, web_recv_chan):
        # 没有 web_gui, 接受全部数据转发给下游 api_recv_chan
        async for pack in web_recv_chan:
            await api_recv_chan.send(pack)

    async def _data_handler(self, api_recv_chan, web_recv_chan):
        async for pack in web_recv_chan:
            if pack['aid'] == 'rtn_data':
                web_diffs = []
                account_changed = False
                for d in pack['data']:
                    # 把 d 处理成需要的数据
                    # 处理 trade
                    trade = d.get("trade")
                    if trade is not None:
                        TqWebHelper.merge_diff(self._data["trade"], trade)
                        web_diffs.append({"trade": trade})
                        # 账户是否有变化
                        static_balance_changed = d.get("trade", {}).get(self._api._account._account_id, {}).\
                            get("accounts", {}).get("CNY", {}).get('static_balance')
                        trades_changed = d.get("trade", {}).get(self._api._account._account_id, {}).get("trades", {})
                        orders_changed = d.get("trade", {}).get(self._api._account._account_id, {}).get("orders", {})
                        if static_balance_changed is not None or trades_changed != {} or orders_changed != {}:
                            account_changed = True
                    # 处理 backtest
                    tqsdk_backtest = d.get("_tqsdk_backtest")
                    if tqsdk_backtest is not None:
                        TqWebHelper.merge_diff(self._data, d)
                        web_diffs.append(d)
                        if self._data["action"]["mode"] != "backtest":
                            TqWebHelper.merge_diff(self._data, {"action": {"mode": "backtest"}})
                            web_diffs.append({"action": {"mode": "backtest"}})
                    # 处理通知，行情和交易连接的状态
                    notify_diffs = self._notify_handler(d.get("notify", {}))
                    for diff in notify_diffs:
                        TqWebHelper.merge_diff(self._data, diff)
                    web_diffs.extend(notify_diffs)
                if account_changed:
                    dt, snapshot = self.get_snapshot()
                    _snapshots = {"snapshots": {}}
                    _snapshots["snapshots"][dt] = snapshot
                    web_diffs.append(_snapshots)
                    TqWebHelper.merge_diff(self._data, _snapshots)
                for chan in self._conn_diff_chans:
                    self.send_to_conn_chan(chan, web_diffs)
            # 接收的数据转发给下游 api
            await api_recv_chan.send(pack)

    def _notify_handler(self, notifies):
        """将连接状态的通知转成 diff 协议"""
        diffs = []
        for _, notify in notifies.items():
            if notify["code"] == 2019112901 or notify["code"] == 2019112902:
                # 连接建立的通知 第一次建立 或者 重连建立
                if notify["url"] == self._api._md_url:
                    diffs.append({
                        "action": {
                            "md_url_status": True
                        }
                    })
                elif notify["url"] == self._api._td_url:
                    diffs.append({
                        "action": {
                            "td_url_status": True
                        }
                    })
            elif notify["code"] == 2019112911:
                # 连接断开的通知
                if notify["url"] == self._api._md_url:
                    diffs.append({
                        "action": {
                            "md_url_status": False
                        }
                    })
                elif notify["url"] == self._api._td_url:
                    diffs.append({
                        "action": {
                            "td_url_status": False
                        }
                    })
        return diffs

    def send_to_conn_chan(self, chan, diffs):
        last_diff = chan.recv_latest({})
        for d in diffs:
            TqWebHelper.merge_diff(last_diff, d, reduce_diff = False)
        if last_diff != {}:
            chan.send_nowait(last_diff)

    def dt_func (self):
        if '_tqsdk_backtest' in self._data:
            return self._data['_tqsdk_backtest']['current_dt']
        else:
            return int(datetime.now().timestamp() * 1e9)

    def get_snapshot(self):
        account = self._data.get("trade", {}).get(self._api._account._account_id, {}).get("accounts", {}).get("CNY", {})
        positions = self._data.get("trade", {}).get(self._api._account._account_id, {}).get("positions", {})
        dt = self.dt_func()
        return dt, {
            'accounts': {'CNY': {k: v for k, v in account.items() if not k.startswith("_")}},
            'positions': {k: {pk: pv for pk, pv in v.items() if not pk.startswith("_")} for k, v in
                          positions.items() if
                          not k.startswith("_")}
        }

    @staticmethod
    def get_obj(root, path, default=None):
        """获取业务数据"""
        d = root
        for i in range(len(path)):
            if path[i] not in d:
                dv = {} if i != len(path) - 1 or default is None else default
                d[path[i]] = dv
            d = d[path[i]]
        return d

    @staticmethod
    def merge_diff(result, diff, reduce_diff = True):
        """
        更新业务数据
        :param result: 更新结果
        :param diff: diff pack 
        :param reduce_diff: 表示是否修改 diff 对象本身，因为如果 merge_diff 的 result 是 conn_chan 内部的 last_diff，那么 diff 会在循环中多次使用，这时候一定不能修改 diff 本身
        :return: 
        """
        for key in list(diff.keys()):
            if diff[key] is None:
                result.pop(key, None)
            elif isinstance(diff[key], dict):
                target = TqWebHelper.get_obj(result, [key])
                TqWebHelper.merge_diff(target, diff[key], reduce_diff = reduce_diff)
                if len(diff[key]) == 0:
                    del diff[key]
            elif reduce_diff and key in result and result[key] == diff[key]:
                del diff[key]
            else:
                result[key] = diff[key]

    async def link_wsserver(self):
        async def lambda_connection_handler(conn, path): await self.connection_handler(conn)
        async with websockets.serve(lambda_connection_handler, host='127.0.0.1', port=0) as server:
            port = server.server.sockets[0].getsockname()[1]
            await self.web_port_chan.send({'port': port})
            await asyncio.sleep(100000000000)

    def get_send_msg(self, data=None):
        return simplejson.dumps({
            'aid': 'rtn_data',
            'data': [self._data if data is None else data]
        }, ignore_nan=True)

    async def connection_handler(self, conn):
        send_msg = self.get_send_msg(self._data)
        await conn.send(send_msg)
        conn_chan = tqsdk.api.TqChan(self._api, last_only=True)
        self._conn_diff_chans.add(conn_chan)
        try:
            async for msg in conn:
                pack = simplejson.loads(msg)
                if pack["aid"] == 'peek_message':
                    last_diff = await conn_chan.recv()
                    send_msg = self.get_send_msg(last_diff)
                    await conn.send(send_msg)
        except Exception as e:
            await conn_chan.close()
            self._conn_diff_chans.remove(conn_chan)

    async def link_httpserver(self):

        ws_port = await self.web_port_chan.recv()
        # init http server handlers
        ins_url = self._api._ins_url
        md_url = self._api._md_url
        ws_url = 'ws://127.0.0.1:' + str(ws_port['port'])
        app = web.Application()
        app.router.add_get(path='/url',
                           handler=lambda request: TqWebHelper.httpserver_url_handler(ins_url, md_url, ws_url))
        app.router.add_get(path='/', handler=lambda request: TqWebHelper.httpserver_index_handler(self._web_dir))
        app.router.add_static('/', self._web_dir, show_index=True)
        runner = web.AppRunner(app)
        await runner.setup()
        server_socket = socket.socket()
        server_socket.bind(('127.0.0.1', self._http_server_port))
        address = server_socket.getsockname()
        site = web.SockSite(runner, server_socket)
        await site.start()
        self._logger.info("您可以访问 http://{ip}:{port} 查看策略绘制出的 K 线图形。".format(ip=address[0], port=address[1]))
        await asyncio.sleep(100000000000)

    @staticmethod
    def httpserver_url_handler(ins_url, md_url, ws_url):
        return web.json_response({
                'ins_url': ins_url,
                'md_url': md_url,
                'ws_url': ws_url
            })

    @staticmethod
    def httpserver_index_handler(web_dir):
        return web.FileResponse(path=web_dir + '/index.html')
